"""Exceptions used by the shleem package."""


class ConfigurationException(Exception):  # pragma: no cover
    """An exception thrown when a problem is found in a configuration mapping
    used to configure shleem."""
    pass
